#
# Copyright 2013 Metamarkets Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
class Postaggregator:
    def __init__(self, fn, fields, name):
        self.post_aggregator = {
            "type": "arithmetic",
            "name": name,
            "fn": fn,
            "fields": fields,
        }
        self.name = name

    def __mul__(self, other):
        return Postaggregator("*", self.fields(other), self.name + "mul" + other.name)

    def __sub__(self, other):
        return Postaggregator("-", self.fields(other), self.name + "sub" + other.name)

    def __add__(self, other):
        return Postaggregator("+", self.fields(other), self.name + "add" + other.name)

    def __div__(self, other):
        return Postaggregator("/", self.fields(other), self.name + "div" + other.name)

    def __truediv__(self, other):
        return self.__div__(other)

    def fields(self, other):
        return [self.post_aggregator, other.post_aggregator]

    @staticmethod
    def build_post_aggregators(postaggs):
        def rename_postagg(new_name, post_aggregator):
            post_aggregator["name"] = new_name
            return post_aggregator

        return [
            rename_postagg(new_name, postagg.post_aggregator)
            for (new_name, postagg) in postaggs.items()
        ]


class Quantile(Postaggregator):
    def __init__(self, name, probability):
        Postaggregator.__init__(self, None, None, name)
        self.post_aggregator = {
            "type": "quantile",
            "fieldName": name,
            "probability": probability,
        }


class Quantiles(Postaggregator):
    def __init__(self, name, probabilities):
        Postaggregator.__init__(self, None, None, name)
        self.post_aggregator = {
            "type": "quantiles",
            "fieldName": name,
            "probabilities": probabilities,
        }


class Field(Postaggregator):
    def __init__(self, name):
        Postaggregator.__init__(self, None, None, name)
        self.post_aggregator = {"type": "fieldAccess", "fieldName": name}


class Const(Postaggregator):
    def __init__(self, value, output_name=None):

        if output_name is None:
            name = "const"
        else:
            name = output_name

        Postaggregator.__init__(self, None, None, name)
        self.post_aggregator = {"type": "constant", "name": name, "value": value}


class HyperUniqueCardinality(Postaggregator):
    def __init__(self, name):
        Postaggregator.__init__(self, None, None, name)
        self.post_aggregator = {"type": "hyperUniqueCardinality", "fieldName": name}


class DoubleGreatest(Postaggregator):
    def __init__(self, fields, output_name=None):

        if output_name is None:
            name = "doubleGreatest"
        else:
            name = output_name

        Postaggregator.__init__(self, None, None, name)
        self.post_aggregator = {
            "type": "doubleGreatest",
            "name": name,
            "fields": [f.post_aggregator for f in fields],
        }


class DoubleLeast(Postaggregator):
    def __init__(self, fields, output_name=None):

        if output_name is None:
            name = "doubleLeast"
        else:
            name = output_name

        Postaggregator.__init__(self, None, None, name)
        self.post_aggregator = {
            "type": "doubleLeast",
            "name": name,
            "fields": [f.post_aggregator for f in fields],
        }


class LongGreatest(Postaggregator):
    def __init__(self, fields, output_name=None):

        if output_name is None:
            name = "longGreatest"
        else:
            name = output_name

        Postaggregator.__init__(self, None, None, name)
        self.post_aggregator = {
            "type": "longGreatest",
            "name": name,
            "fields": [f.post_aggregator for f in fields],
        }


class LongLeast(Postaggregator):
    def __init__(self, fields, output_name=None):

        if output_name is None:
            name = "longLeast"
        else:
            name = output_name

        Postaggregator.__init__(self, None, None, name)
        self.post_aggregator = {
            "type": "longLeast",
            "name": name,
            "fields": [f.post_aggregator for f in fields],
        }


class ThetaSketchOp(object):
    def __init__(self, fn, fields, name):
        self.post_aggregator = {
            "type": "thetaSketchSetOp",
            "name": name,
            "func": fn,
            "fields": fields,
        }
        self.name = name

    def __or__(self, other):
        return ThetaSketchOp(
            "UNION", self.fields(other), self.name + "_OR_" + other.name
        )

    def __and__(self, other):
        return ThetaSketchOp(
            "INTERSECT", self.fields(other), self.name + "_AND_" + other.name
        )

    def __ne__(self, other):
        return ThetaSketchOp(
            "NOT", self.fields(other), self.name + "_NOT_" + other.name
        )

    def fields(self, other):
        return [self.post_aggregator, other.post_aggregator]

    @staticmethod
    def build_post_aggregators(thetasketchops):
        def rename_thetasketchop(new_name, thetasketchop):
            thetasketchop["name"] = new_name
            return thetasketchop

        return [
            rename_thetasketchop(new_name, thetasketchop.post_aggregator)
            for (new_name, thetasketchop) in thetasketchops.items()
        ]


class ThetaSketch(ThetaSketchOp):
    def __init__(self, name):
        ThetaSketchOp.__init__(self, None, None, name)
        self.post_aggregator = {"type": "fieldAccess", "fieldName": name}


class ThetaSketchEstimate(Postaggregator):
    def __init__(self, fields):
        field = (
            fields.post_aggregator
            if type(fields) in [ThetaSketch, ThetaSketchOp]
            else fields
        )
        self.post_aggregator = {
            "type": "thetaSketchEstimate",
            "name": "thetasketchestimate",
            "field": field,
        }
        self.name = "thetasketchestimate"



#
# Copyright 2013 Metamarkets Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
try:
    import simplejson as json
except ImportError:
    import json


class Having:
    def __init__(self, **args):

        if args["type"] in ("equalTo", "lessThan", "greaterThan"):
            self.having = {
                "having": {
                    "type": args["type"],
                    "aggregation": args["aggregation"],
                    "value": args["value"],
                }
            }

        elif args["type"] == "and":
            self.having = {
                "having": {"type": "and", "havingSpecs": args["havingSpecs"]}
            }

        elif args["type"] == "or":
            self.having = {"having": {"type": "or", "havingSpecs": args["havingSpecs"]}}

        elif args["type"] == "not":
            self.having = {"having": {"type": "not", "havingSpec": args["havingSpec"]}}

        elif args["type"] == "filter":
            self.having = {"having": {"type": args["type"], "filter": args["filter"]}}

        elif args["type"] == "dimSelector":
            self.having = {
                "having": {
                    "type": args["type"],
                    "dimension": args["dimension"],
                    "value": args["value"],
                }
            }

        else:
            raise NotImplementedError(
                "Having type: {0} does not exist".format(args["type"])
            )

    def show(self):
        print(json.dumps(self.having, indent=4))

    def _combine(self, typ, x):
        # collapse nested and/ors
        if self.having["having"]["type"] == typ:
            havingSpecs = self.having["having"]["havingSpecs"] + [x.having["having"]]
            return Having(type=typ, havingSpecs=havingSpecs)
        elif x.having["having"]["type"] == typ:
            havingSpecs = [self.having["having"]] + x.having["having"]["havingSpecs"]
            return Having(type=typ, havingSpecs=havingSpecs)
        else:
            return Having(
                type=typ, havingSpecs=[self.having["having"], x.having["having"]]
            )

    def __and__(self, x):
        return self._combine("and", x)

    def __or__(self, x):
        return self._combine("or", x)

    def __invert__(self):
        return Having(type="not", havingSpec=self.having["having"])

    @staticmethod
    def build_having(having_obj):
        return having_obj.having["having"]


class Aggregation:
    def __init__(self, agg):
        self.aggregation = agg

    def __eq__(self, other):
        return Having(type="equalTo", aggregation=self.aggregation, value=other)

    def __lt__(self, other):
        return Having(type="lessThan", aggregation=self.aggregation, value=other)

    def __gt__(self, other):
        return Having(type="greaterThan", aggregation=self.aggregation, value=other)


class Dimension:
    def __init__(self, dim):
        self.dimension = dim

    def __eq__(self, other):
        return Having(type="dimSelector", dimension=self.dimension, value=other)



import os
import re
import sys
from urllib import parse

from prompt_toolkit import prompt
from prompt_toolkit.completion.word_completer import WordCompleter
from prompt_toolkit.history import FileHistory
from pygments.lexers import SqlLexer
from pygments.style import Style
from pygments.styles.default import DefaultStyle
from pygments.token import Token
from tabulate import tabulate

from pydruid.db.api import connect

keywords = [
    "EXPLAIN PLAN FOR",
    "WITH",
    "SELECT",
    "ALL",
    "DISTINCT",
    "FROM",
    "WHERE",
    "GROUP BY",
    "HAVING",
    "ORDER BY",
    "ASC",
    "DESC",
    "LIMIT",
    "FILTER",
    "UNION ALL",
]

aggregate_functions = [
    "COUNT",
    "SUM",
    "MIN",
    "MAX",
    "AVG",
    "APPROX_COUNT_DISTINCT",
    "APPROX_QUANTILE",
]

numeric_functions = [
    "ABS",
    "CEIL",
    "EXP",
    "FLOOR",
    "LN",
    "LOG10",
    "POWER",
    "SQRT",
    "TRUNCATE",
    "TRUNC",
    "MOD",
]

string_functions = [
    "LENGTH",
    "CHAR_LENGTH",
    "CHARACTER_LENGTH",
    "STRLEN",
    "LOOKUP",
    "LOWER",
    "REGEXP_EXTRACT",
    "REPLACE",
    "STRPOS",
    "SUBSTRING",
    "SUBSTR",
    "TRIM",
    "BTRIM",
    "RTRIM",
    "LTRIM",
    "UPPER",
]

time_functions = [
    "CURRENT_TIMESTAMP",
    "CURRENT_DATE",
    "DATE_TRUNC",
    "TIME_FLOOR",
    "TIME_SHIFT",
    "TIME_EXTRACT",
    "TIME_PARSE",
    "TIME_FORMAT",
    "MILLIS_TO_TIMESTAMP",
    "TIMESTAMP_TO_MILLIS",
    "EXTRACT",
    "FLOOR",
    "CEIL",
    "TIMESTAMPADD",
]

other_functions = ["CAST", "CASE", "WHEN", "THEN", "END", "NULLIF", "COALESCE"]


replacements = {
    "^SHOW SCHEMAS": "SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA",
    "^SHOW TABLES": "SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES",
    "^DESC (?P<table>[^;\\s]*)": r"""
        SELECT COLUMN_NAME,
               ORDINAL_POSITION,
               COLUMN_DEFAULT,
               IS_NULLABLE,
               DATA_TYPE
          FROM INFORMATION_SCHEMA.COLUMNS
         WHERE TABLE_NAME='\1'
    """.strip(),
}


class DocumentStyle(Style):
    styles = {
        Token.Menu.Completions.Completion.Current: "bg:#00aaaa #000000",
        Token.Menu.Completions.Completion: "bg:#008888 #ffffff",
        Token.Menu.Completions.ProgressButton: "bg:#003333",
        Token.Menu.Completions.ProgressBar: "bg:#00aaaa",
    }
    styles.update(DefaultStyle.styles)


def get_connection_kwargs(url):
    parts = parse.urlparse(url)
    if ":" in parts.netloc:
        host, port = parts.netloc.split(":", 1)
        port = int(port)
    else:
        host = parts.netloc
        port = 8082

    return {"host": host, "port": port, "path": parts.path, "scheme": parts.scheme}


def get_tables(connection):
    cursor = connection.cursor()
    return [
        row.TABLE_NAME
        for row in cursor.execute("SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES")
    ]


def get_autocomplete(connection):
    return (
        keywords
        + aggregate_functions
        + numeric_functions
        + string_functions
        + time_functions
        + other_functions
        + get_tables(connection)
    )


def main():
    history = FileHistory(os.path.expanduser("~/.pydruid_history"))

    try:
        url = sys.argv[1]
    except IndexError:
        url = "http://localhost:8082/druid/v2/sql/"
    kwargs = get_connection_kwargs(url)
    connection = connect(**kwargs)
    cursor = connection.cursor()

    words = get_autocomplete(connection)
    sql_completer = WordCompleter(words, ignore_case=True)

    while True:
        try:
            query = prompt(
                "> ",
                lexer=SqlLexer,
                completer=sql_completer,
                style=DocumentStyle,
                history=history,
            )
        except (EOFError, KeyboardInterrupt):
            break  # Control-D pressed.

        # run query
        query = query.strip("; ")
        if query:
            # shortcuts
            if query.lower() in ("bye", "exit", "quit"):
                break
            for pattern, repl in replacements.items():
                query = re.sub(pattern, repl, query)

            try:
                result = cursor.execute(query)
            except Exception as e:
                print(e)
                continue

            headers = [t[0] for t in cursor.description or []]
            print(tabulate(result, headers=headers))

    print("GoodBye!")


if __name__ == "__main__":
    main()
